/*
G. Wiesenekker. GWD a draughts program.
Copyright (C) 1993  G. Wiesenekker
E-mail: wiesenekker@sara.nl

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <signal.h>

#define MAC    0
#define DOS    1
#define RS6000 2

#ifdef __MSDOS__
#define MACHINE DOS
#endif

#ifndef MACHINE
#define MACHINE MAC
#endif

#define FALSE 0
#define TRUE 1
#define INVALID (-1)
#define LINE_MAX 256

#define and &&
#define or  ||

#define ABS(A) ((A) > 0 ? (A) : -(A))
#define MAX(A, B) ((A) > (B) ? (A) : (B))
#define BIT(X)  (1 << (X))
#define BITL(X) (1L << (X))

#define FATAL_ERROR(X)\
	{my_printf("%s::%lu:%s\n", __FILE__, (ulong_t) __LINE__, X); exit(EXIT_FAILURE);}
	
#define BUG(X) {if (X) FATAL_ERROR(#X)}
#define MALLOC(P, T, N) BUG(((P) = (T *) malloc(sizeof(T) * (N))) == NULL)

typedef signed char byte_t;
typedef unsigned long ulong_t;

#define NGEN_MAX   256

#define WHITE_MAN  0
#define BLACK_MAN  1
#define NOTHING    2
#define ANYTHING   3
#define EMPTY      4

char *code[EMPTY] = {"WM", "BM", "EMP", "ANY"};

#define NAME_MAX 31

#define PAT_FUNC 0
#define PAT_DEF  1

typedef struct
{
	byte_t pat_board[50];
	char pat_name[NAME_MAX + 1];
	int pat_ready;
} pat_t;

void my_printf(char *, ...);
void my_printf(char *format, ...)
{
	va_list arg;
#if MACHINE == MAC
	static FILE *stdecho = NULL;
#endif
	
	va_start(arg, format);
	vfprintf(stdout, format, arg);
	fflush(stdout);
#if MACHINE == MAC
	if (stdecho == NULL)
		BUG((stdecho = fopen("echo", "w")) == NULL)
	vfprintf(stdecho, format, arg);
	fflush(stdecho);
#endif
	va_end(arg);
}

FILE *fcode;
int *freq[EMPTY];
int ndebug;

void gen_code(char *, int, pat_t **, pat_t **);
void gen_code(char *tab, int ngen, pat_t **gen, pat_t **temp)
{
	int ikind, iboard, jkind, jboard, igen;
	int freq_max;
	pat_t *with;
	int ntemp;
	
	while(TRUE)
	{
		/*
		If the number of remaining fields is 0, the pattern is in the
		enclosing if's. Add the score.
		*/
		
		for (igen = 0; igen < ngen; igen++)
		{
			with = gen[igen];
			if (with->pat_ready) continue;
			
			ntemp = 0;
			for (iboard = 0; iboard < 50; iboard++)
				if (with->pat_board[iboard] != EMPTY) ntemp++;
				
			/*
			If the number of remaining fields is 0, the pattern is in the
			enclosing if's. Add the score.
			If the number of remaining fields is 1, add the 'if' and the score.
			If the number of remaining fields is greate than 1, add
			'if and' and the score.
			*/
			
			if (ntemp == 0)
			{
				fprintf(fcode, "%ss += %s;\n", tab, with->pat_name);
				with->pat_ready = TRUE;
				ndebug++;
			}
		}
		
		/*
		Clear frequencies.
		*/
		
		for (ikind = 0; ikind < EMPTY; ikind++)
			for (iboard = 0; iboard < 50; iboard++) freq[ikind][iboard] = 0;
		
		/*
		Find maximum frequency.
		*/
		
		jkind = jboard = INVALID;
		freq_max = 0;
		for (igen = 0; igen < ngen; igen++)
		{
			with = gen[igen];
			if (with->pat_ready) continue;
			
			for (iboard = 0; iboard < 50; iboard++)
			{
				ikind = with->pat_board[iboard];
				if ((ikind != EMPTY) and (ikind != INVALID))
				{
					++(freq[ikind][iboard]);
					if (freq[ikind][iboard] > freq_max)
					{
						jkind = ikind;
						jboard = iboard;
						freq_max = freq[ikind][iboard];
					}
				}
			}
		}
		
		/*
		Done if the maximum frequency is 0
		*/
		
		if (freq_max == 0) break;
		
		BUG(jkind == INVALID)
		BUG(jboard == INVALID)
		
		/*
		If the maximum frequency is 1, the patterns have no common field.
		*/
		
		if (freq_max == 1)
		{
			for (igen = 0; igen < ngen; igen++)
			{
				with = gen[igen];
				if (with->pat_ready) continue;
				
				/*
				Find the number of remaining fields in the pattern.
				*/
				
				ntemp = 0;
				for (iboard = 0; iboard < 50; iboard++)
					if (with->pat_board[iboard] != EMPTY) ntemp++;
				BUG(ntemp == 0)
				
				/*
				If the number of remaining fields is 1, add the 'if' and the score.
				If the number of remaining fields is greate than 1, add
				'if and' and the score.
				*/
				
				if (ntemp == 1)
				{
					jboard = 0;
					while(with->pat_board[jboard] == EMPTY) jboard++;
					jkind = with->pat_board[jboard];
					fprintf(fcode, "%sif (%s(%d)) s += %s;\n",
						tab, code[jkind], jboard + 1, with->pat_name);
					with->pat_board[jboard] = EMPTY;
					with->pat_ready = TRUE;
					ndebug++;
				}
				else
				{
					jboard = 0;
					while(with->pat_board[jboard] == EMPTY) jboard++;
					jkind = with->pat_board[jboard];
					fprintf(fcode, "%sif (%s(%d)", tab, code[jkind], jboard + 1);
					with->pat_board[jboard] = EMPTY;
					for (iboard = jboard + 1; iboard < 50; iboard++)
					{
						if (with->pat_board[iboard] != EMPTY)
						{
							ikind = with->pat_board[iboard];
							fprintf(fcode, " and %s(%d)", code[ikind], iboard + 1);
							with->pat_board[iboard] = EMPTY;
						}
					}
					fprintf(fcode, ") s += %s;\n", with->pat_name);
					with->pat_ready = TRUE;
					ndebug++;
				}
			}
		}
		else
		{
			fprintf(fcode, "%sif (%s(%d))\n%s{\n", tab, code[jkind], jboard + 1, tab);
			ntemp = 0;
			for (igen = 0; igen < ngen; igen++)
			{
				with = gen[igen];
				if (with->pat_board[jboard] == jkind)
				{
					with->pat_board[jboard] = EMPTY;
					temp[ntemp++] = with;
				}
			}
			BUG(ntemp == 0)
			gen_code(tab - 1, ntemp, temp, temp + ntemp);
			fprintf(fcode, "%s}\n", tab);
		}
	}
}

int main(void);
int main(void)
{
	int itab;
	char tab[51];
	FILE *f;
	int ngen;
	pat_t *pat;
	pat_t **gen, **temp;
	pat_t *with;
	int ikind, iboard;
	char line[LINE_MAX];
	size_t len;
	
	for (itab = 0; itab < 50; itab++)
		tab[itab] = '\t';
	tab[50] = '\0';
	
	for (ikind = 0; ikind < EMPTY; ikind++)
		MALLOC(freq[ikind], int, 50)
	
	MALLOC(pat, pat_t, NGEN_MAX)
	MALLOC(gen, pat_t *, NGEN_MAX)
	MALLOC(temp, pat_t *, NGEN_MAX)
	BUG((fcode = fopen("pat.c", "w")) == NULL)
	fprintf(fcode, "#include \"draughts.h\"\n");
	fprintf(fcode, "#include \"pat.h\"\n");
	BUG((f = fopen("gencode.dat", "r")) == NULL)
	ngen = 0;
	while(TRUE)
	{
		BUG(ngen >= NGEN_MAX)
		with = pat + ngen;
		gen[ngen] = with;
		if (fscanf(f, "%s", line) != 1) goto label_ready;
		len = strlen(line);
		BUG(len == 0);
		strncpy(with->pat_name, line, NAME_MAX);
/*
#if MACHINE == MAC
		if (line[len - 1] == ')')
		{
			line[len - 2] = '\0';
			fprintf(fcode, "int %s(void);\nint %s(void){return(13);}\n", line, line);
		}
		else
			fprintf(fcode, "#define %s 13\n", line);
#endif
*/
		for (iboard = 0; iboard < 50; iboard++)
		{
			BUG(fscanf(f, "%s", line) != 1)
			if (strcmp(line, "wO") == 0)
				with->pat_board[iboard] = WHITE_MAN;
			else if (strcmp(line, "bO") == 0)
				with->pat_board[iboard] = BLACK_MAN;
			else if (strcmp(line, "xx") == 0)
				with->pat_board[iboard] = NOTHING;
			else if (strcmp(line, "??") == 0)
				with->pat_board[iboard] = ANYTHING;
			else if (strcmp(line, "..") == 0)
				with->pat_board[iboard] = EMPTY;
			else
			{
				my_printf("line=%s\n", line);
				FATAL_ERROR("unknown piece")
			}
		}
		with->pat_ready = FALSE;
		ngen++;
	}
	
	label_ready:
	BUG(fclose(f) == EOF)
	my_printf("ngen=%d\n", ngen);
	fprintf(fcode, "int return_pat_score(void)\n{\n");
	fprintf(fcode, "register int s = 0;\n");
	ndebug = 0;
	gen_code(tab + 50, ngen, gen, temp);
	BUG(ndebug != ngen)
	fprintf(fcode, "return(s);\n}\n");
	BUG(fclose(fcode) == EOF)
	return(EXIT_SUCCESS);
}
