/*
	G. Wiesenekker. ZZZZZZ a chess program.
	Copyright (C) 1993-1994 G. Wiesenekker
	E-mail: wiesenekker@sara.nl

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _ZZZZZZ_H
#define _ZZZZZZ_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <math.h>
#include <signal.h>
#include <ctype.h>

/*
Machine dependencies are treated by #if MACHINE == 
ZZZZZZ knows only about MAC, DOS, SGI and SUN. That is, #define MACHINE RS6000 or
CRAY are all equivalent.
*/

#define MAC    0
#define DOS    1
#define RS6000 2
#define CRAY   3
#define SGI    4
#define SUN    5

#ifdef __MSDOS__
	#include <alloc.h>
	#define MACHINE DOS
#endif

#ifndef MACHINE
	#define MACHINE MAC
#endif

#if MACHINE == SUN
	#define EXIT_SUCCESS 0
	#define EXIT_FAILURE 1
	#define CLOCKS_PER_SEC 1000000L
	#define difftime(X, Y) ((X) - (Y))
#endif

#define VERSION_STRING_MAX 32

/*
Setting DEBUG_LEVEL to 1 includes debugging code in the program.
Makes it run slow, but it is very useful if you run into problems.
*/

#undef DEBUG_LEVEL

/*
Defining SMALL reduces the sizes of some tables used by zzzzzz.
Useful when debugging under DOS.
*/

#undef SMALL

/*
DepthMax is the maximum search depth.
TimeLimit is the allocated time.
GameTimeUsed is the total time used for the game.
If Append is TRUE we may append moves to the game file.
If UseBook is TRUE we may use the opening book.
If UsePos is TRUE we use the static evaluation.
*/

extern int DepthMax;
extern int TimeLimit;
extern int GameTimeUsed;
extern int MateDepth;
extern int Append;
extern int UseBook;
extern int UsePos;
#ifdef PVM
extern int PvmProcs;
extern int PvmDepth;
extern int PvmLegal;
#endif

/*
Some general macro's used by every C programmer.
*/

#define FALSE 0
#define TRUE 1
#define INVALID (-1)

/*
Cancel rediculous predefined values of LINE_MAX like 65535
*/

#undef LINE_MAX
#define LINE_MAX 256

/*
I have had some very nasty bugs in the past by typing & instead of &&.
I think that the word static is overused in C.
*/

#define and &&
#define or  ||
#define local static

#define ABS(A) ((A) > 0 ? (A) : -(A))
#define MAX(A, B) ((A) > (B) ? (A) : (B))
#define BIT(X)  (1 << (X))
#define BITL(X) (1L << (X))
#define BITUL(X) (1UL << (X))

/*
Leave out the error string in DOS to save precious memory.
*/

#if MACHINE == DOS
	#define FATAL_ERROR(X)\
		{my_printf("??%s::%lu\n", FILE_NAME, (ulong_t) __LINE__); exit(EXIT_FAILURE);}
#else
	#define FATAL_ERROR(X)\
		{my_printf("%s::%lu:%s\n", FILE_NAME, (ulong_t) __LINE__, X); exit(EXIT_FAILURE);}
#endif

/*
You will find many BUG macro's in the code. Of course, the error messages
are only understandable if you look at the code.
*/

#define BUG(X) {if (X) FATAL_ERROR(#X)}

#if MACHINE == DOS
	#define MALLOC(P, T, N)\
	{\
		my_printf("malloc %s size=%lu number=%lu total=%lu\n", #P, (ulong_t) sizeof(T), (ulong_t) (N), (ulong_t) sizeof(T) * (N));\
		BUG(((P) = (T *) malloc(sizeof(T) * (N))) == NULL)\
	}
	#define FREE(P) {free(P);}
#else
	#define MALLOC(P, T, N) BUG(((P) = (T *) malloc(sizeof(T) * (N))) == NULL)
	#define FREE(P) {free(P);}
#endif
#define FCLOSE(F) BUG(fclose(F) == EOF)
#define FREAD(F, P, T, N) BUG(fread(P, sizeof(T), N, F) != N)
#define FREAD_STRING(F, P) {long N; FREAD(F, &N, long, 1); FREAD(F, P, char, N); }
#define FWRITE(F, P, T, N) BUG(fwrite(P, sizeof(T), N, F) != N)
#define FWRITE_STRING(F, P) {long N; N = strlen(P) + 1; FWRITE(F, &N, long, 1); FWRITE(F, P, char, N); }
#define FSEEK(F, N) BUG(fseek(F, N, SEEK_SET) != 0)
#define FBOF(F) BUG(fseek(F, 0, SEEK_SET) != 0)
#define FEOF(F) BUG(fseek(F, 0, SEEK_END) != 0)
#define FTELL(F, X) BUG(((X) = ftell(F)) < 0)
#if MACHINE == DOS
	#define HUGE_MALLOC(P, T, N) BUG(((P) = (T huge *) farmalloc(sizeof(T) * (N))) == NULL)
	#define HUGE_FREE(P) {farfree((void far *) P);}
	#define HUGE_FREAD(P, T, N, F)\
	{\
		long i;\
		T t;\
		for (i = 0; i < (N); i++)\
		{\
			BUG(fread(&t, sizeof(T), 1, F) != 1)\
			(P)[i] = t;\
		}\
	}
	#define HUGE_FWRITE(P, T, N, F)\
	{\
		long i;\
		T t;\
		for (i = 0; i < (N); i++)\
		{\
			t = (P)[i];\
			BUG(fwrite(&t, sizeof(T), 1, F) != 1)\
		}\
	}
#endif

/*
DUMMY disables some compiler warnings about unused variables or arguments
(unused now, but may be used in the future).
dummy is always FALSE, but the compiler does not know that.
*/

#define DUMMY(F, V) if (dummy) my_printf(F, V);
extern int dummy;

typedef signed char byte_t;
#if ((MACHINE != SGI) and (MACHINE != RS6000))
typedef unsigned char uchar_t;
typedef unsigned int uint_t;
typedef unsigned long ulong_t;
#endif

/*
WHITE_COLOUR and BLACK_COLOUR to avoid name clashes with WHITE and BLACK under DOS.
*/

enum {WHITE_COLOUR = 0, BLACK_COLOUR, NONE, SIDE};
extern char *colour_as_string[2];

/*
The value of the pieces, which are the normal chess values
1, 3, 3, 5, 10 scaled by a factor of 300 (to allow for some
flexibility in the evaluation function parameters).
The knight and bishop score are 3.5 pawn so that the program
does not exchange too easily a knight or bishop for three pawns.

*/

#define PAWN_SCORE   (3 * 100)
#define KNIGHT_SCORE (3 * 350)
#define BISHOP_SCORE (3 * 350)
#define ROOK_SCORE   (3 * 550)
#define QUEEN_SCORE  (3 * 1100)
#define KING_SCORE   0

/*
The following macro's define bits for the different kind of moves
in chess.
*/

#define NULL_MOVE   BITL(16)
#define PROM_CAPT   BITL(17)
#define PROM_NORM   BITL(18)
#define CAPT        BITL(19)
#define EP          BITL(20)
#define SHORT       BITL(21)
#define LONG        BITL(22)
#define DOUBLE      BITL(23)
#define NORM        BITL(24)
#define PROM_KNIGHT BITL(25)
#define PROM_BISHOP BITL(26)
#define PROM_ROOK   BITL(27)
#define PROM_QUEEN  BITL(28)
#define INVAL_MOVE  BITL(29)

typedef int score_t;

/*
In a future version I will squeeze the move in an int (6 bits from,
6 bits to, 4 bits kind).
*/

typedef long move_t;

/*
We use 64 bit (or 128 in case of problems) hash keys.
See T.A. Marsland, ICCA vol. 9, no. 1 (1986) pp. 3-19.
T. Warnock and B. Wendroff, ICCA vol. 11, no 1 (1988), pp. 10-12
*/

#undef HASH_KEY128

typedef struct
{
	ulong_t low, high;
#ifdef HASH_KEY128
	ulong_t lower, higher;
#endif
} hash_key_t;

#ifdef HASH_KEY128
	#define CLEAR_KEY(X)\
		{(X).low = (X).high = (X).lower = (X).higher = 0UL;}
	#define UPDATE_KEY(X, Y)\
		{(X).low ^= (Y).low; (X).high ^= (Y).high; (X).lower ^= (Y).lower; (X).higher ^= (Y).higher;}
	#define EQUAL_KEY(X, Y)\
		(((X).low == (Y).low) and ((X).high == (Y).high) and ((X).lower == (Y).lower) and ((X).higher == (Y).higher))
	#define SET_KEY(X)\
		{(X).low = ulrand(); (X).high = ulrand(); (X).lower = ulrand(); (X).higher = ulrand();}
#else
	#define CLEAR_KEY(X)\
		{(X).low = (X).high = 0UL;}
	#define UPDATE_KEY(X, Y)\
		{(X).low ^= (Y).low; (X).high ^= (Y).high;}
	#define EQUAL_KEY(X, Y)\
		(((X).low == (Y).low) and ((X).high == (Y).high))
	#define SET_KEY(X)\
		{(X).low = ulrand(); (X).high = ulrand();}
#endif

/*
We store the pieces in lists, one list for every piece, so that
we can first generate the pawn moves, then the knight moves etc.
(Note: I usually prefix the names of the fields in a struct with
the name of the struct).
board_colour is WHITE_COLOUR, BLACK_COLOUR, SIDE or NONE.
board_kind is INVALID, PAWN, KNIGHT, ROOK, QUEEN or KING.
If board_kind is INVALID, the end of the list is reached.
board_value is the material value of the piece.
board_where tells where the piece is on the board.
if board_where is INVALID, the piece has been captured.
So you will see often this loop:

board_t *p;
(loop until end of list)
for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
{
	(ignore captured pieces)
	if (p->board_where == INVALID) continue;
	..
}
board_moved tells whether the piece has moved (useful for
checking castling rights and pieces that have not been
developed).
board_captured tells whether the piece has captured (useful
in the quiescence search).
board_key is a pointer to a table with hash keys.
board_dir is a pointer to a table of offsets for the 
different directions in which a piece can move.
board_shift tells whether we may offset the piece more than
once (FALSE for PAWN, KNIGHT and KING, TRUE for BISHOP,
ROOK and QUEEN).
board_count is a pointer to a table with piece counts.
*/

#define LIST_MAX 11

typedef struct
{
	int board_colour, board_kind, board_value, board_where;
	int board_moved, board_captured;
#if MACHINE == DOS
	hash_key_t huge *board_key;
#else
	hash_key_t *board_key;
#endif
	int board_shift;
	int *board_dir;
	int *board_count;
} board_t;

/*
my_colour is the colour to move. my_colour and your_colour are toggled by
do_move and undo_move.
*/
#define WHITE_TO_MOVE {my_colour = WHITE_COLOUR; your_colour = BLACK_COLOUR;}
#define BLACK_TO_MOVE {my_colour = BLACK_COLOUR; your_colour = WHITE_COLOUR;}

extern int my_colour, your_colour;

/*
A node contains all kinds of information about a node.
We need capt and prom for undoing captures and promotions.
We need node_moved and node_captured for resetting the
moved and captured attributes when undoing a move.
node_mat contains the current material balance for WHITE and BLACK.
node_key contains the current hash key.
node_total_key contains the total hash key (including ep or 
castling rights).
node_reverse tells whether the move was reversible (useful for
checking for draw by repetition).
node_hack_ep contains the square where we may take en passant (will
be INVALID most of the time).
node_move is the move itself.
*/

typedef struct
{
	board_t *capt, *prom;
	int node_moved, node_captured;
	hash_key_t node_key, node_total_key;
	int node_mat[2];
	int node_reverse;
	int node_hack_ep;
	move_t node_move;
} node_t;

/*
tot_count contains the total piece counts for white and black.
*/

extern int tot_count[2];

extern node_t *node;

#define SCORE_INVAL   32000

/*
ZZZZZZ knows different move formats. long algebraic notation (e.g. g1f3)
or standard algebraic notation (e.g. Nf3).
*/

#define MOVE_STRING_LONG 0
#define MOVE_STRING_SAN  1
#define MOVE_STRING_ANY  2

#define MOVE_STRING_MAX 16

/*
The legal moves in the starting position are stored in array root.
root_move is the move.
root_score is the move score (not very useful, because it is a lower
and upper bounds for most moves).
*/

typedef struct
{
	move_t root_move;
	score_t root_score;
} root_t;

#define NROOT_MAX 256

/*
var_t is used for printing the best variation.
*/

typedef move_t var_t;

/*
capt_top is a pointer to a stack where we store the moves.
*/

extern move_t *capt_top;

extern char *start_fen;

extern int nroot;
extern root_t *root;
extern int jroot;
extern int GameTimeUsed;

#define DBASE_LINE_MAX 256
extern char dbase_line[DBASE_LINE_MAX];
extern int ndbase_records;
extern char *dbase_records[DBASE_LINE_MAX];

typedef struct
{
	int random_opening;
	int print_dbase;
	int print_distrib;
} options_t;

extern options_t options;

/* 
Prototypes
*/

void zzzzzz_version(char *);
void util_version(char *);
void pgn_version(char *);
char *time_stamp(void);
void my_printf(char *, ...);
char *my_fgets(char *, int, FILE *);
ulong_t ulrand(void);
ulong_t crc(uchar_t *);
void handle_signal(int);
extern char move_string[MOVE_STRING_MAX];
void move_to_string(move_t);
void display_board(void);
void board_to_fen(char *, int);
void fen_to_board(char *, int);
void write_board(FILE *f);
void display_score(void);
void display_list(void);
void display_occup(void);
void display_attack(void);
void compact_lists(void);
int i_am_check(int, int);
int gen_all_moves(void);
void do_move(move_t);
void undo_move(void);
void set_score(double *);
void move_to_san_string(move_t, int);
int search_san(char *);
int draw_by_repetition(int);
void gen_root(void);
int search_root(char *, int, int, int);
int get_book_move(void);
int do_mate(void);
void do_root(void);
void sort_root(void);
void initialize(void);
void pgn(void);
int search_dbase(void);
void dbase(char *);
void btree(void);

#ifdef PVM

	#define PVM_INVALID_TAG  1
	#define PVM_SLAVE_TAG    2
	#define PVM_NUMBER_TAG   3
	#define PVM_ALLOCATE_TAG 4
	#define PVM_UPDATE_TAG   5
	#define PVM_IDLE_TAG     6
	#define PVM_WAITING_TAG  7
	#define PVM_QUIT_TAG     8
	#define PVM_FAMILYID_MIN 9

	#define TRACEPVMBUG(X) {int info = (X); my_printf("%s::%lu:%s\n", FILE_NAME, (ulong_t) __LINE__, #X); if (info < 0) {pvm_perror(#X); FATAL_ERROR("pvm error")}}
	#define PVMBUG(X) {int info = (X); if (info < 0) {pvm_perror(#X); FATAL_ERROR("pvm error")}}

	#define PVM_PROCS_MAX    256
	#define PVM_FAMILY_MAX   PVM_PROCS_MAX
	#define PVM_BROTHERS_MAX 10
	
	void do_master(void);
	void do_driver(void);
	void do_slave(void);
	extern int drivertid;
	#if MACHINE == MAC
		#define PvmTaskDefault 0
		#define PvmDataDefault 0
		
		int pvm_mytid(void);
		int pvm_spawn(char *, char **, int, char *, int, int *);
		int pvm_initsend(int);
		int pvm_pkint(int *, int, int);
		int pvm_pklong(long *, int, int);
		int pvm_pkfloat(float *, int, int);
		int pvm_pkdouble(double *, int, int);
		int pvm_pkstr(char *);
		int pvm_send(int, int);
		int pvm_recv(int, int);
		int pvm_nrecv(int, int);
		int pvm_upkint(int *, int, int);
		int pvm_upklong(long *, int, int);
		int pvm_upkfloat(float *, int, int);
		int pvm_upkdouble(double *, int, int);
		int pvm_upkstr(char *);
		int pvm_exit(void);
		int pvm_parent(void);
		int pvm_mcast(int *, int, int);
		int	pvm_bufinfo(int, int *, int *, int *);
		int	pvm_freebuf(int);
		int pvm_perror(char *);
	#else
		#include "pvm3.h"
	#endif
#endif

#endif
